/// @file particle_pp.cc
/// This example demonstrates loading, running and scripting a very simple NaCl
/// module.  To load the NaCl module, the browser first looks for the
/// CreateModule() factory method (at the end of this file).  It calls
/// CreateModule() once to load the module code from your .nexe.  After the
/// .nexe code is loaded, CreateModule() is not called again.
///
/// Once the .nexe code is loaded, the browser than calls the CreateInstance()
/// method on the object returned by CreateModule().  It calls CreateInstance()
/// each time it encounters an <embed> tag that references your NaCl module.
///
/// When the browser encounters JavaScript that references your NaCl module, it
/// calls the GetInstanceObject() method on the object returned from
/// CreateInstance().  In this example, the returned object is a subclass of
/// ScriptableObject, which handles the scripting support.

// ev[gpi_generator̃R[h؂\B`ɃXbhgȂ悤ɕύX
#include <stdio.h>
#include "particle_pp.h"

namespace {
const char* const kPaintMethodId = "paint";
const char* const kInitMethodId = "init";
	// This is called by the brower when the 2D context has been flushed to the
	// browser window.
	void FlushCallback(void* data, int32_t result) {
	  static_cast<particle_pp::ParticlePpInstance*>(data)->set_flush_pending(false);
	}
}  // namespace

namespace particle_pp {
	class ParticlePpScriptableObject : public pp::deprecated::ScriptableObject {
	private:
		typedef pp::Var (ParticlePpScriptableObject::*Method)(const pp::Var& method, const std::vector<pp::Var>& args, pp::Var* exception);
		ParticlePpInstance	*app_instance_;
		std::map<std::string, Method>	method_table;
	public:
		ParticlePpScriptableObject(ParticlePpInstance* app_instance) : app_instance_(app_instance)
		{
			method_table.insert( std::pair<std::string, Method>( std::string(kPaintMethodId), &ParticlePpScriptableObject::Paint ) );
			method_table.insert( std::pair<std::string, Method>( std::string(kInitMethodId), &ParticlePpScriptableObject::Init ) );
		}
		virtual bool HasMethod(const pp::Var& method, pp::Var* exception);

		virtual pp::Var Call(const pp::Var& method, const std::vector<pp::Var>& args, pp::Var* exception);
		pp::Var Paint(const pp::Var& method, const std::vector<pp::Var>& args, pp::Var* exception);
		pp::Var Init(const pp::Var& method, const std::vector<pp::Var>& args, pp::Var* exception);
	};

	bool ParticlePpScriptableObject::HasMethod(const pp::Var& method, pp::Var* exception)
	{
		if (!method.is_string()) {
			return false;
		}
		std::string method_name = method.AsString();
		std::map<std::string, Method>::iterator it = method_table.find(method_name);
		return it != method_table.end();
	}


	pp::Var ParticlePpScriptableObject::Call(const pp::Var& method,
	                                       const std::vector<pp::Var>& args,
	                                       pp::Var* exception)
	{
		if (!method.is_string()) {
			return pp::Var();
		}
		std::string method_name = method.AsString();
		std::map<std::string, Method>::iterator it = method_table.find(method_name);
		if (it != method_table.end()) {
			return (this->*(it->second))(method, args, exception);
		}
		return pp::Var();
	}

	pp::Var ParticlePpScriptableObject::Paint(const pp::Var& method, const std::vector<pp::Var>& args, pp::Var* exception)
	{
		if (args.size() == 5) {
			double x = args[0].AsDouble();
			double y = args[1].AsDouble();
			double force = args[2].AsDouble();
			bool colorTransform = args[3].AsBool();
			bool use_sse2 = args[4].AsBool();

			if (use_sse2)	app_instance_->updateParticlesSSE2(x,y,force, colorTransform);
			else			app_instance_->updateParticles(x,y,force, colorTransform);

			return app_instance_->Paint();
		}
		return pp::Var(4.0);
	}

	pp::Var ParticlePpScriptableObject::Init(const pp::Var& method, const std::vector<pp::Var>& args, pp::Var* exception)
	{
		if (args.size() == 1) {
			return app_instance_->ParticleInit( args[0].AsInt() );
		}
		return pp::Var(3.0);
	}

	ParticlePpInstance::ParticlePpInstance(PP_Instance instance) 
	:	pp::Instance(instance),
		graphics_2d_context_(NULL),
		pixel_buffer_(NULL),
		flush_pending_(false)

	{
	}

	ParticlePpInstance::~ParticlePpInstance()
	{
		DestroyContext();

		delete pixel_buffer_;
	}
	

	void ParticlePpInstance::DidChangeView(const pp::Rect& position,  const pp::Rect& clip)
	{
		if (position.size().width() == width() &&
		  position.size().height() == height())
			return;  // Size didn't change, no need to update anything.

		// Create a new device context with the new size.
		DestroyContext();
		CreateContext(position.size());
		delete pixel_buffer_;	pixel_buffer_ = NULL;
		
		if (graphics_2d_context_ != NULL) {
			pixel_buffer_ = new pp::ImageData(this, PP_IMAGEDATAFORMAT_BGRA_PREMUL, graphics_2d_context_->size(), false);
		}
	}
	pp::Var ParticlePpInstance::GetInstanceObject()
	{
		ParticlePpScriptableObject* hw_object = new ParticlePpScriptableObject(this);
		return pp::Var(this, hw_object);
	}

	bool ParticlePpInstance::Init(uint32_t argc, const char* argn[], const char* argv[])
	{
		return true;
	}
	

	pp::Var ParticlePpInstance::Paint()
	{
		return FlushPixelBuffer();
	}
	

	void ParticlePpInstance::CreateContext(const pp::Size& size)
	{
		if (IsContextValid()) return;
		
		graphics_2d_context_ = new pp::Graphics2D(this, size, false);
		
		if (!BindGraphics(*graphics_2d_context_)) {
			printf("Couldn't bind the device context\n");
		}
	}

	void ParticlePpInstance::DestroyContext()
	{
		if (!IsContextValid()) return;
		
		delete graphics_2d_context_;
		graphics_2d_context_ = NULL;
	}

	pp::Var ParticlePpInstance::FlushPixelBuffer() {
		if (!IsContextValid())	return pp::Var(0.0);
		graphics_2d_context_->PaintImageData(*pixel_buffer_, pp::Point());
		if (flush_pending())	return pp::Var(0.0);

		set_flush_pending(true);
		graphics_2d_context_->Flush(pp::CompletionCallback(&FlushCallback, this));
		return pp::Var(1.0);
	}

	pp::Var ParticlePpInstance::ParticleInit(int count)
	{
		delete particles;

		particle_count = count;
		particles = new std::vector<float>(count*8);
		int index = 0;
		for (int i = 0; i < particle_count; i++) {
			(*particles)[index++] = (RANDOM()*465);
			(*particles)[index++] = (RANDOM()*465);
			(*particles)[index++] = 0;
			(*particles)[index++] = 0;
			(*particles)[index++] = (RANDOM()*100+127);
			(*particles)[index++] = (RANDOM()*100+127);
			(*particles)[index++] = (RANDOM()*128+127);
			double r = RANDOM();
			if (RANDOM() < 0.8) {
				(*particles)[index++] = r < 1/16.0 ? 1/16.0+RANDOM()/32.0 : r < 1/8.0 ? 1/8.0+RANDOM()/16.0 : r < 1/4.0 ? 1/4.0+RANDOM()/8.0 : 1/2.0;
			} else {
				(*particles)[index++] = r < 1/2.0 ? r : r/2.0;
			}
		}
		return true;
	}

	pp::Var ParticlePpInstance::updateParticlesSSE2(float mx, float my, float force, bool colorTransform)
	{
		return pp::Var(2.0);
	}
	
	pp::Var ParticlePpInstance::updateParticles(float mx, float my, float force, bool colorTransform)
	{
		if (!particles) return pp::Var(0.0);
		
		uint32_t* pixel_bits =  reinterpret_cast<uint32_t*>(pixel_buffer_->data());
		if (pixel_bits == NULL) return pp::Var(0.0);

		double gravPoint_x = mx + RANDOM()*16;
		double gravPoint_y = my + RANDOM()*16;
		
		int w = width();
		int h = height();
		
		if (colorTransform) {
			uint32_t* p = pixel_bits;
			for (int y = 0; y < h; y++) {
				for (int x = 0; x < w; x++) {
					union rgb32 col;
					col.rgba = *p;
					if (force > 0) {
						col.b.r *= 0.7;
						col.b.g *= 0.8;
						col.b.b *= 0.96;
					} else {
						col.b.r *= 0.96;
						col.b.g *= 0.8;
						col.b.b *= 0.7;
					}
					*p++ = col.rgba | 0xff000000;
				}
			}
		} else {
			memset(pixel_bits, 0, w*h*sizeof(uint32_t));
			uint32_t* p = pixel_bits;
			for (int y = 0; y < h; y++) {
				for (int x = 0; x < w; x++) {
					*p++ = 0xff000000;
				}
			}
		}

		int count = particle_count;
		int index = 0;
		for (int i = 0; i < count; i++) {
			double px = (*particles)[index+0];
			double py = (*particles)[index+1];
			double vx = (*particles)[index+2];
			double vy = (*particles)[index+3];
			
			double diff_x = gravPoint_x-px;
			double diff_y = gravPoint_y-py;
			double acc = (diff_x*diff_x + diff_y*diff_y);
			if (acc != 0.0) acc = force/acc; else acc = 0.0;
			double acc_x = diff_x*acc;
			double acc_y = diff_y*acc;
			vx += acc_x - acc_y*(*particles)[index+7];
			vy += acc_y + acc_x*(*particles)[index+7];
			
			px += vx;
			py += vy;
			
			vx *= 0.96;
			vy *= 0.96;
			
			if (px >= w)	px -= w;
			if (px < 0)		px += w;
			if (py >= h)	py -= h;
			if (py < 0)		py += h;

			int r = (int)(*particles)[index+4];
			int g = (int)(*particles)[index+5];
			int b = (int)(*particles)[index+6];
			
			int ix = (int)px; ix = ix >= w ? w-1 : ix < 0 ? 0 : ix;
			int iy = (int)py; iy = iy >= h ? h-1 : iy < 0 ? 0 : iy;
			pixel_bits[w*iy + ix] = (r << 16) | (g << 8) | b | 0xff000000;

			(*particles)[index+0] = px;
			(*particles)[index+1] = py;
			(*particles)[index+2] = vx;
			(*particles)[index+3] = vy;
			
			index += 8;
		}
		
		return pp::Var(1.0);
	}

}

